from typing import List, Iterator, Dict
from overrides import overrides
from antu.io.token_indexers.token_indexer import TokenIndexer
from antu.io.vocabulary import Vocabulary
from antu.io.fields.field import Field


class TextField(Field):
    """
    A ``TextField`` is a data field that is commonly used in NLP tasks, and we
    can use it to store text sequences such as sentences, paragraphs, POS tags,
    and so on.

    Parameters
    ----------
    name : ``str``
        Field name. This is necessary and must be unique (not the same as other
        field names).
    tokens : ``List[str]``
        Field content that contains a list of string.
    indexers : ``List[TokenIndexer]``, optional (default=``list()``)
        Indexer list that defines the vocabularies associated with the field.
    """
    def __init__(
        self,
        name: str,
        tokens: List[str],
        indexers: List[TokenIndexer] = list()):
        self.name = name
        self.tokens = tokens
        self.indexers = indexers

    def __iter__(self) -> Iterator[str]:
        return iter(self.tokens)

    def __getitem__(self, idx: int) -> str:
        return self.tokens[idx]

    def __len__(self) -> int:
        return len(self.tokens)

    @overrides
    def count_vocab_items(
        self,
        counters: Dict[str, Dict[str, int]]) -> None:
        """
        We count the number of strings if the string needs to be counted to some
         counters. You can pass directly if there is no string that needs
        to be counted.

        Parameters
        ----------
        counters : ``Dict[str, Dict[str, int]]``
            Element statistics for datasets. if field indexers indicate that
            this field is related to some counters, we use field content to
            update the counters.
        """
        for idxer in self.indexers:
            for token in self.tokens:
                idxer.count_vocab_items(token, counters)

    @overrides
    def index(
        self,
        vocab: Vocabulary) -> None:
        """
        Gets one or more index mappings for each element in the Field.

        Parameters
        ----------
        vocab : ``Vocabulary``
            ``vocab`` is used to get the index of each item.
        """
        self.indexes = {}
        for idxer in self.indexers:
            self.indexes.update(idxer.tokens_to_indices(self.tokens, vocab))


from typing import Dict, Optional, Union, List, Set, TypeVar
from bidict import bidict

DEFAULT_PAD_TOKEN = "*@PAD@*"
DEFAULT_UNK_TOKEN = "*@UNK@*"

class Vocabulary(object):
    """
    Parameters
    ----------
    counters : ``Dict[str, Dict[str, int]]``, optional (default= ``dict()`` )
        Element statistics for datasets.
    min_count : ``Dict[str, int]``, optional (default= ``dict()`` )
        Defines the minimum number of occurrences when some counter are
        converted to vocabulary.
    pretrained_vocab : ``Dict[str, List[str]]``, optional (default=  ``dict()``
        External pre-trained vocabulary.
    intersection_vocab : ``Dict[str, str]``, optional (default= ``dict()`` )
        Defines the intersection with which vocabulary takes, when loading some
        oversized pre-trained vocabulary.
    no_pad_namespace : ``Set[str]``, optional (default= ``set()`` )
        Defines which vocabularies do not have `pad` token.
    no_unk_namespace : ``Set[str]``, optional (default= ``set()`` )
        Defines which vocabularies do not have `oov` token.
    """

    def __init__(self,
                 counters: Dict[str, Dict[str, int]] = dict(),
                 min_count: Dict[str, int] = dict(),
                 pretrained_vocab: Dict[str, List[str]] = dict(),
                 intersection_vocab: Dict[str, str] = dict(),
                 no_pad_namespace: Set[str] = set(),
                 no_unk_namespace: Set[str] = set()):

        self._PAD_token = DEFAULT_PAD_TOKEN
        self._UNK_token = DEFAULT_UNK_TOKEN
        self.min_count = min_count
        self.intersection_vocab = intersection_vocab
        self.no_unk_namespace = no_unk_namespace
        self.no_pad_namespace = no_pad_namespace
        self.vocab_cnt = {}
        self.vocab = {}

        for vocab_name, counter in dict(counters, **pretrained_vocab).items():
            self.vocab[vocab_name] = bidict()
            cnt = 0

            # Handle unknown token
            if vocab_name not in no_unk_namespace:
                self.vocab[vocab_name][self._UNK_token] = cnt
                cnt += 1

            # Handle padding token
            if vocab_name not in no_pad_namespace:
                self.vocab[vocab_name][self._PAD_token] = cnt
                cnt += 1

            # Build Vocabulary from Dataset Counter
            if isinstance(counter, dict):
                minn = (min_count[vocab_name]
                        if min_count and vocab_name in min_count else 0)
                for key, value in counter.items():
                    if value >= minn:
                        self.vocab[vocab_name][key] = cnt
                        cnt += 1

            # Build Vocabulary from Pretrained Vocabulary List
            elif isinstance(counter, list):
                is_intersection = vocab_name in intersection_vocab
                target_vocab = (self.vocab[intersection_vocab[vocab_name]]
                                if is_intersection else {})
                for key in counter:
                    if not is_intersection or key in target_vocab:
                        self.vocab[vocab_name][key] = cnt
                        cnt += 1
            self.vocab_cnt[vocab_name] = cnt


    def extend_from_pretrained_vocab(
        self,
        pretrained_vocab: Dict[str, List[str]],
        intersection_vocab: Dict[str, str] = dict(),
        no_pad_namespace: Set[str] = set(),
        no_unk_namespace: Set[str] = set()) -> None:
        """
        Extend the vocabulary from the pre-trained vocabulary after defining
        the vocabulary.

        Parameters
        ----------
        pretrained_vocab : ``Dict[str, List[str]]``
            External pre-trained vocabulary.
        intersection_vocab : ``Dict[str, str]``, optional (default= ``dict()`` )
            Defines the intersection with which vocabulary takes, when loading some
            oversized pre-trained vocabulary.
        no_pad_namespace : ``Set[str]``, optional (default= ``set()`` )
            Defines which vocabularies do not have `pad` token.
        no_unk_namespace : ``Set[str]``, optional (default= ``set()`` )
            Defines which vocabularies do not have `oov` token.
        """
        self.no_unk_namespace.update(no_unk_namespace)
        self.no_pad_namespace.update(no_pad_namespace)
        self.intersection_vocab.update(intersection_vocab)
        for vocab_name, counter in pretrained_vocab.items():
            print(vocab_name)
            self.vocab[vocab_name] = bidict()

            cnt = 0
            # Handle unknown token
            if vocab_name not in no_unk_namespace:
                self.vocab[vocab_name][self._UNK_token] = cnt
                cnt += 1

            # Handle padding token
            if vocab_name not in no_pad_namespace:
                self.vocab[vocab_name][self._PAD_token] = cnt
                cnt += 1

            # Build Vocabulary from Pretrained Vocabulary List
            is_intersection = vocab_name in intersection_vocab
            target_vocab = (self.vocab[intersection_vocab[vocab_name]]
                            if is_intersection else {})
            for key in counter:
                if not is_intersection or key in target_vocab:
                    self.vocab[vocab_name][key] = cnt
                    cnt += 1
            self.vocab_cnt[vocab_name] = cnt

    def extend_from_counter(
        self,
        counters: Dict[str, Dict[str, int]],
        min_count: Union[int, Dict[str, int]] = dict(),
        no_pad_namespace: Set[str] = set(),
        no_unk_namespace: Set[str] = set()) -> None:
        """
        Extend the vocabulary from the dataset statistic counters after defining
        the vocabulary.

        Parameters
        ----------
        counters : ``Dict[str, Dict[str, int]]``
            Element statistics for datasets.
        min_count : ``Dict[str, int]``, optional (default= ``dict()`` )
            Defines the minimum number of occurrences when some counter are
            converted to vocabulary.
        no_pad_namespace : ``Set[str]``, optional (default= ``set()`` )
            Defines which vocabularies do not have `pad` token.
        no_unk_namespace : ``Set[str]``, optional (default= ``set()`` )
            Defines which vocabularies do not have `oov` token.
        """
        self.no_unk_namespace.update(no_unk_namespace)
        self.no_pad_namespace.update(no_pad_namespace)
        self.min_count.update(min_count)

        for vocab_name, counter in counters.items():
            self.vocab[vocab_name] = bidict()
            cnt = 0
            # Handle unknown token
            if vocab_name not in no_unk_namespace:
                self.vocab[vocab_name][self._UNK_token] = cnt
                cnt += 1

            # Handle padding token
            if vocab_name not in no_pad_namespace:
                self.vocab[vocab_name][self._PAD_token] = cnt
                cnt += 1

            # Build Vocabulary from Dataset Counter
            minn = (min_count[vocab_name]
                    if min_count and vocab_name in min_count else 0)
            for key, value in counter.items():
                if value >= minn:
                    self.vocab[vocab_name][key] = cnt
                    cnt += 1
            self.vocab_cnt[vocab_name] = cnt

    def add_token_to_namespace(self, token: str, namespace: str) -> None:
        """
        Extend the vocabulary by add token to vocabulary namespace.

        Parameters
        ----------
        token : ``str``
            The token that needs to be added.
        namespace : ``str``
            Which vocabulary needs to be added to.
        """
        self.vocab[namespace][token] = self.vocab_cnt[namespace]
        self.vocab_cnt[namespace] += 1

    def get_token_index(self, token: str, vocab_name: str) -> int:
        """
        Gets the index of a token in the vocabulary.

        Parameters
        ----------
        token : ``str``
            Gets the index of which token.
        namespace : ``str``
            Which vocabulary this token belongs to.

        Returns
        -------
        Index : ``int``
        """
        if token in self.vocab[vocab_name]:
            return self.vocab[vocab_name][token]
        elif vocab_name not in self.no_unk_namespace:
            return self.vocab[vocab_name][self._UNK_token]
        else:
            raise RuntimeError(
                'Try to get a OOV token (%s)\'s index from a no unknown token '
                'vocabulary (%s)' % (token, vocab_name))

    def get_token_from_index(self, index: int, vocab_name: str) -> str:
        """
        Gets the token of a index in the vocabulary.

        Parameters
        ----------
        index : ``int``
            Gets the token of which index.
        namespace : ``str``
            Which vocabulary this index belongs to.

        Returns
        -------
        Token : ``str``
        """
        if index < self.vocab_cnt[vocab_name]:
            return self.vocab[vocab_name].inv[index]
        else:
            raise RuntimeError(
                'Index (%d) out of vocabulary (%s) range'
                % (index, vocab_name))

    def get_vocab_size(self, namespace: str) -> int:
        """
        Gets the size of a vocabulary.

        Parameters
        ----------
        namespace : ``str``
            Which vocabulary.

        Returns
        -------
        Vocabulary size : ``int``
        """
        return len(self.vocab[vocab_name])







from typing import Dict, List, Callable, TypeVar
from overrides import overrides
from antu.io.vocabulary import Vocabulary
from antu.io.token_indexers.token_indexer import TokenIndexer
Indices = TypeVar("Indices", List[int], List[List[int]])

class CharTokenIndexer(TokenIndexer):
    """
    A ``CharTokenIndexer`` determines how string token get represented as
    arrays of list of character indices in a model.

    Parameters
    ----------
    related_vocabs : ``List[str]``
        Which vocabularies are related to the indexer.
    transform : ``Callable[[str,], str]``, optional (default=``lambda x:x``)
        What changes need to be made to the token when counting or indexing.
        Commonly used are lowercase transformation functions.
    """
    def __init__(
        self,
        related_vocabs: List[str],
        transform: Callable[[str,], str]=lambda x:x) -> None:
        self.related_vocabs = related_vocabs
        self.transform = transform

    @overrides
    def count_vocab_items(
        self,
        token: str,
        counters: Dict[str, Dict[str, int]]) -> None:
        """
        Each character in the token is counted directly as an element.

        Parameters
        ----------
        counter : ``Dict[str, Dict[str, int]]``
            We count the number of strings if the string needs to be counted to
            some counters.
        """
        for vocab_name in self.related_vocabs:
            if vocab_name in counters:
                for ch in token:
                    counters[vocab_name][self.transform(ch)] += 1

    @overrides
    def tokens_to_indices(
        self,
        tokens: List[str],
        vocab: Vocabulary) -> Dict[str, List[List[int]]]:
        """
        Takes a list of tokens and converts them to one or more sets of indices.
        During the indexing process, each token item corresponds to a list of
        index in the vocabulary.

        Parameters
        ----------
        vocab : ``Vocabulary``
            ``vocab`` is used to get the index of each item.
        """
        res = {}
        for vocab_name in self.related_vocabs:
            index_list = []

            for token in tokens:
                index_list.append(
                    [vocab.get_token_index(self.transform(ch), vocab_name)
                    for ch in token])
            res[vocab_name] = index_list
        return res




import pytest
from antu.io.vocabulary import Vocabulary
from collections import Counter


class TestVocabulary:

    def test_extend_from_pretrained_vocab(self):
        vocab = Vocabulary()

        # Test extend a vocabulary from a simple pretained vocab
        pretrained_vocabs = {'glove': ['a', 'b', 'c']}
        vocab.extend_from_pretrained_vocab(pretrained_vocabs)
        assert vocab.get_token_index('a', 'glove') == 2
        assert vocab.get_token_index('c', 'glove') == 4
        assert vocab.get_token_index('d', 'glove') == 0

        # Test extend a vocabulary from a pretained vocabulary,
        # and intersect with another vocabulary.
        pretrained_vocabs = {'w2v': ['b', 'c', 'd']}
        vocab.extend_from_pretrained_vocab(pretrained_vocabs, {'w2v':'glove'})
        assert vocab.get_token_index('b', 'w2v') == 2
        assert vocab.get_token_index('d', 'w2v') == 0
        assert vocab.get_token_from_index(2, 'w2v') == 'b'
        with pytest.raises(RuntimeError) as excinfo:
            vocab.get_token_from_index(4, 'w2v')
        assert excinfo.type == RuntimeError

        # Test extend a vocabulary from a no oov pretained vocabulary
        pretrained_vocabs = {'glove_nounk': ['a', 'b', 'c']}
        vocab.extend_from_pretrained_vocab(
            pretrained_vocabs, no_unk_namespace={'glove_nounk',})
        assert vocab.get_token_index('a', 'glove_nounk') == 1
        assert vocab.get_token_index('c', 'glove_nounk') == 3
        with pytest.raises(RuntimeError) as excinfo:
            vocab.get_token_index('d', 'glove_nounk')
        assert excinfo.type == RuntimeError

        # Test extend a vocabulary from a no oov and pad pretained vocabulary
        pretrained_vocabs = {'glove_nounk_nopad': ['a', 'b', 'c']}
        vocab.extend_from_pretrained_vocab(
            pretrained_vocabs,
            no_unk_namespace={'glove_nounk_nopad',},
            no_pad_namespace={"glove_nounk_nopad"})
        assert vocab.get_token_index('a', 'glove_nounk_nopad') == 0
        assert vocab.get_token_index('c', 'glove_nounk_nopad') == 2
        with pytest.raises(RuntimeError) as excinfo:
            vocab.get_token_index('d', 'glove_nounk_nopad')
        assert excinfo.type == RuntimeError

    def test_extend_from_counter(self):
        vocab = Vocabulary()

        # Test extend a vocabulary from a simple counter
        counter = {'w': Counter(["This", "is", "a", "test", "sentence", '.'])}
        vocab.extend_from_counter(counter)
        assert vocab.get_token_index('a', 'w') == 4
        assert vocab.get_token_index('.', 'w') == 7
        assert vocab.get_token_index('That', 'w') == 0

        # Test extend a vocabulary from a counter with min_count
        counter = {'w_m': Counter(['This', 'is', 'is'])}
        min_count = {'w_m': 2}
        vocab.extend_from_counter(counter, min_count)
        assert vocab.get_token_index('is', 'w_m') == 2
        assert vocab.get_token_index('This', 'w_m') == 0
        assert vocab.get_token_index('That', 'w_m') == 0

        # Test extend a vocabulary from a counter without oov token
        counter = {'w_nounk': Counter(['This', 'is'])}
        vocab.extend_from_counter(counter, no_unk_namespace={'w_nounk',})
        with pytest.raises(RuntimeError) as excinfo:
            vocab.get_token_index('That', 'w_nounk')
        assert excinfo.type == RuntimeError
        assert vocab.get_token_index('This', 'w_nounk') == 1

        # Test extend a vocabulary from a counter without pad & unk token
        counter = {'w_nounk_nopad': Counter(['This', 'is', 'a'])}
        vocab.extend_from_counter(
            counter,
            no_unk_namespace={'w_nounk_nopad'},
            no_pad_namespace={'w_nounk_nopad'})
        with pytest.raises(RuntimeError) as excinfo:
            vocab.get_token_index('That', 'w_nounk_nopad')
        assert excinfo.type == RuntimeError
        assert vocab.get_token_index('This', 'w_nounk_nopad') == 0

    def test_vocabulary(self):
        pretrained_vocabs = {
            'glove': ['a', 'b', 'c'],
            'w2v': ['b', 'c', 'd'],
            'glove_nounk': ['a', 'b', 'c'],
            'glove_nounk_nopad': ['a', 'b', 'c']}

        counters = {
            'w': Counter(["This", "is", "a", "test", "sentence", '.']),
            'w_m': Counter(['This', 'is', 'is']),
            'w_nounk': Counter(['This', 'is']),
            'w_nounk_nopad': Counter(['This', 'is', 'a'])}

        vocab = Vocabulary(
            counters=counters,
            min_count={'w_m': 2},
            pretrained_vocab=pretrained_vocabs,
            intersection_vocab={'w2v': 'glove'},
            no_pad_namespace={'glove_nounk_nopad', 'w_nounk_nopad'},
            no_unk_namespace={
                'glove_nounk', 'w_nounk', 'glove_nounk_nopad', 'w_nounk_nopad'})

        # Test glove
        print(vocab.get_vocab_size('glove'))
        assert vocab.get_token_index('a', 'glove') == 2
        assert vocab.get_token_index('c', 'glove') == 4
        assert vocab.get_token_index('d', 'glove') == 0

        # Test w2v
        assert vocab.get_token_index('b', 'w2v') == 2
        assert vocab.get_token_index('d', 'w2v') == 0
        assert vocab.get_token_from_index(2, 'w2v') == 'b'
        with pytest.raises(RuntimeError) as excinfo:
            vocab.get_token_from_index(4, 'w2v')
        assert excinfo.type == RuntimeError

        # Test glove_nounk
        assert vocab.get_token_index('a', 'glove_nounk') == 1
        assert vocab.get_token_index('c', 'glove_nounk') == 3
        with pytest.raises(RuntimeError) as excinfo:
            vocab.get_token_index('d', 'glove_nounk')
        assert excinfo.type == RuntimeError

        # Test glove_nounk_nopad
        assert vocab.get_token_index('a', 'glove_nounk_nopad') == 0
        assert vocab.get_token_index('c', 'glove_nounk_nopad') == 2
        with pytest.raises(RuntimeError) as excinfo:
            vocab.get_token_index('d', 'glove_nounk_nopad')
        assert excinfo.type == RuntimeError

        # Test w
        assert vocab.get_token_index('a', 'w') == 4
        assert vocab.get_token_index('.', 'w') == 7
        assert vocab.get_token_index('That', 'w') == 0

        # Test w_m
        assert vocab.get_token_index('is', 'w_m') == 2
        assert vocab.get_token_index('This', 'w_m') == 0
        assert vocab.get_token_index('That', 'w_m') == 0

        # Test w_nounk
        with pytest.raises(RuntimeError) as excinfo:
            vocab.get_token_index('That', 'w_nounk')
        assert excinfo.type == RuntimeError
        assert vocab.get_token_index('This', 'w_nounk') == 1

        # Test w_nounk_nopad
        with pytest.raises(RuntimeError) as excinfo:
            vocab.get_token_index('That', 'w_nounk_nopad')
        assert excinfo.type == RuntimeError
        assert vocab.get_token_index('This', 'w_nounk_nopad') == 0


import pytest
from antu.io.token_indexers.single_id_token_indexer import SingleIdTokenIndexer
from antu.io.fields.text_field import TextField
from collections import Counter
from antu.io.vocabulary import Vocabulary

class TestSingleIdTokenIndexer:

    def test_single_id_token_indexer(self):
        sentence = ['This', 'is', 'is', 'a', 'a', 'test', 'sentence']
        counter = {'my_word': Counter()}
        vocab = Vocabulary()
        glove = ['This', 'is', 'glove', 'sentence', 'vocabulary']
        vocab.extend_from_pretrained_vocab({'glove': glove})
        indexer = SingleIdTokenIndexer(['my_word', 'glove'])
        sent = TextField('sentence', sentence, [indexer])

        # Test count_vocab_items()
        sent.count_vocab_items(counter)
        assert counter['my_word']['This'] == 1
        assert counter['my_word']['is'] == 2
        assert counter['my_word']['That'] == 0

        vocab.extend_from_counter(counter)

        # Test index()
        sent.index(vocab)
        assert sent.indexes['glove'] == [2, 3, 3, 0, 0, 0, 5]
        assert sent.indexes['my_word'] == [2, 3, 3, 4, 4, 5, 6]


from typing import Dict, List
from abc import ABCMeta, abstractmethod
from antu.io.vocabulary import Vocabulary
from antu.io.instance import Instance


class Dataset(metaclass=ABCMeta):

    vocabulary_set: Vocabulary = {}
    datasets: Dict[str, List[Instance]] = {}

    @abstractmethod
    def read():
        pass

    @abstractmethod
    def input_to_instance():
        pass





from typing import List, Dict, Iterator
from overrides import overrides
from antu.io.token_indexers.token_indexer import TokenIndexer
from antu.io.vocabulary import Vocabulary
from antu.io.fields.field import Field


class IndexField(Field):
    """
    A ``IndexField`` is an integer field, and we can use it to store data ID.

    Parameters
    ----------
    name : ``str``
        Field name. This is necessary and must be unique (not the same as other
        field names).
    tokens : ``List[str]``
        Field content that contains a list of string.
    indexers : ``List[TokenIndexer]``, optional (default=``list()``)
        Indexer list that defines the vocabularies associated with the field.
    """
    def __init__(
        self,
        name: str,
        tokens: List[str],
        indexers: List[TokenIndexer] = None):
        self.name = name
        self.tokens = tokens
        self.indexers = indexers

    def __iter__(self) -> Iterator[str]:
        return iter(self.tokens)

    def __getitem__(self, idx: int) -> str:
        return self.tokens[idx]

    def __len__(self) -> int:
        return len(self.tokens)

    @overrides
    def count_vocab_items(
        self,
        counters: Dict[str, Dict[str, int]]) -> None:
        """
        ``IndexField`` doesn't need index operation.
        """
        pass

    @overrides
    def index(
        self,
        vocab: Vocabulary) -> None:
        """
        ``IndexField`` doesn't need index operation.
        """
        pass






"""Attention mechanism module
Attention is most commonly used in many NLP tasks.
It can be used in any sequence model to look back at past states.
We have implemented some of the attention mechanisms used.
"""

import _dynet as dy


class VanillaAttention(object):
    """This computes Additive attention:
    The original attention mechanism (Bahdanau et al., 2015)
    uses a one-hidden layer feed-forward network to calculate the attention alignment:
    .. math::
        \\begin{split}
           f_{att}(\\boldsymbol{h}_i, \\boldsymbol{s}_t) &=
           \\boldsymbol{v}_a{}^\\top
           \\text{tanh}(\\boldsymbol{W}_1 \\boldsymbol{h}_i +
                       \\boldsymbol{W}_2 \\boldsymbol{s}_t) \\\\
           \\boldsymbol{a}_t &= \\text{softmax}(f_{att}(\\boldsymbol{h}_i, \\boldsymbol{s}_t)) \\\\
           \\boldsymbol{c}_t &= \sum_i a_t^i \\boldsymbol{h}_i \\\\
        \end{split}
    :param model dynet.ParameterCollection:
    :param v_dim int: Dimension of :math:`\\boldsymbol{v}`
    :param h_dim int: Dimension of :math:`\\boldsymbol{h}`
    :param s_dim int: Dimension of :math:`\\boldsymbol{s}`
    :returns: attention vector :math:`\\boldsymbol{c}_t`
    :rtype: dynet.Expression
    """
    def __init__(self, model, v_dim, h_dim, s_dim):
        pc = model.add_subcollection()

        self.W1 = pc.add_parameters((v_dim, h_dim))
        self.W2 = pc.add_parameters((v_dim, s_dim))
        self.v  = pc.add_parameters((1, v_dim))

        self.pc = pc
        self.spec = v_dim, h_dim, s_dim

    def __call__(self, s_t, h_matrix):
        e_t = self.v * dy.tanh(self.W1*h_matrix + self.W2 * s_t)
        a_t = dy.softmax(dy.transpose(e_t))
        c_t = h_matrix * a_t
        return c_t

    @staticmethod
    def from_spec(spec, model):
        """Create and return a new instane with the needed parameters.

        It is one of the prerequisites for Dynet save/load method.
        """
        v_dim, h_dim, s_dim = spec
        return VanillaAttention(model, v_dim, h_dim, s_dim)

    def param_collection(self):
        """Return a :code:`dynet.ParameterCollection` object with the parameters.

        It is one of the prerequisites for Dynet save/load method.
        """
        return self.pc


import pytest
from antu.io.token_indexers.single_id_token_indexer import SingleIdTokenIndexer
from antu.io.token_indexers.char_token_indexer import CharTokenIndexer
from antu.io.fields.text_field import TextField
from collections import Counter
from antu.io.vocabulary import Vocabulary
from antu.io.instance import Instance

class TestInstance:

    def test_instance(self):
        sentence = ['This', 'is', 'is', 'a', 'a', 'test', 'sentence']
        counter = {'my_word': Counter(), 'my_char': Counter()}
        vocab = Vocabulary()
        glove = ['This', 'is', 'glove', 'sentence', 'vocabulary']
        vocab.extend_from_pretrained_vocab({'glove': glove})
        single_id = SingleIdTokenIndexer(['my_word', 'glove'])
        char = CharTokenIndexer(['my_char'])
        sent = TextField('sentence', sentence, [single_id, char])
        data = Instance([sent])

        # Test count_vocab_items()
        data.count_vocab_items(counter)
        assert counter['my_word']['This'] == 1
        assert counter['my_word']['is'] == 2
        assert counter['my_word']['That'] == 0
        assert counter['my_char']['s'] == 5
        assert counter['my_char']['T'] == 1
        assert counter['my_char']['t'] == 3
        assert counter['my_char']['A'] == 0

        vocab.extend_from_counter(counter)

        # Test index()
        result = data.index_fields(vocab)
        assert result['sentence']['glove'] == [2, 3, 3, 0, 0, 0, 5]
        assert result['sentence']['my_word'] == [2, 3, 3, 4, 4, 5, 6]
        assert result['sentence']['my_char'][0] == [2, 3, 4, 5] # 'This'
        assert result['sentence']['my_char'][1] == result['sentence']['my_char'][2]
        assert result['sentence']['my_char'][3] == result['sentence']['my_char'][4]


from typing import List, Dict
from abc import ABCMeta, abstractmethod

from antu.io.vocabulary import Vocabulary


class Field(metaclass=ABCMeta):
    """
    A ``Field`` is an ingredient of a data instance. In most NLP tasks, ``Field``
    stores data of string types. It contains one or more indexers that map string
    data to the corresponding index. Data instances are collections of fields.
    """
    @abstractmethod
    def count_vocab_items(self, counter: Dict[str, Dict[str, int]]) -> None:
        """
        We count the number of strings if the string needs to be mapped to one
        or more integers. You can pass directly if there is no string that needs
        to be mapped.

        Parameters
        ----------
        counter : ``Dict[str, Dict[str, int]]``
        ``counter`` is used to count the number of each item. The first key
        represents the namespace of the vocabulary, and the second key represents
        the string of the item.
        """
        pass

    @abstractmethod
    def index(self, vocab: Vocabulary) -> None:
        """
        Gets one or more index mappings for each element in the Field.

        Parameters
        ----------
        vocab : ``Vocabulary``
        ``vocab`` is used to get the index of each item.
        """
        pass



import _dynet as dy

class MLP(object):
    """docstring for MLP"""
    def __init__(self, model, hidden_sizes, act=dy.tanh, bias=True, dropout=0.0):
        pc = model.add_subcollection()
        sizes = hidden_sizes
        self.W = [
            pc.add_parameters((x, y))
            for x, y in zip(sizes[1:], sizes[:-1])]
        self.b = [
            pc.add_parameters((y,), init=dy.ConstInitializer(0))
            for y in sizes[1:]]

        self.pc = pc
        self.act = act
        self.bias = bias
        self.dropout = dropout
        self.spec = (hidden_sizes, act, bias, dropout)

    def __call__(self, x, train=False):
        h = x
        for W, b in zip(self.W, self.b):
            h = self.act(dy.affine_transform([b, W, h]))
            if train:
                h = dy.dropout_dim(h, 1, self.dropout)
        return h

    @staticmethod
    def from_spec(spec, model):
        """Create and return a new instane with the needed parameters.

        It is one of the prerequisites for Dynet save/load method.
        """
        hidden_sizes, act_func, bias, dropout = spec
        return MLP(model, hidden_sizes, act_func, bias, dropout)

    def param_collection(self):
        """Return a :code:`dynet.ParameterCollection` object with the parameters.

        It is one of the prerequisites for Dynet save/load method.
        """
        return self.pc


import pytest
from antu.io.token_indexers.char_token_indexer import CharTokenIndexer
from antu.io.fields.text_field import TextField
from collections import Counter
from antu.io.vocabulary import Vocabulary

class TestCharTokenIndexer:

    def test_char_token_indexer(self):
        sentence = ['This', 'is', 'is', 'a', 'a', 'test', 'sentence']
        counter = {'my_char': Counter()}
        vocab = Vocabulary()
        glove = ['a', 'b', 'c', 'd', 'e']
        vocab.extend_from_pretrained_vocab({'glove': glove})
        indexer = CharTokenIndexer(['my_char', 'glove'])
        sent = TextField('sentence', sentence, [indexer])

        # Test count_vocab_items()
        sent.count_vocab_items(counter)
        assert counter['my_char']['s'] == 5
        assert counter['my_char']['T'] == 1
        assert counter['my_char']['t'] == 3
        assert counter['my_char']['A'] == 0

        vocab.extend_from_counter(counter)

        # Test index()
        sent.index(vocab)
        assert sent.indexes['glove'][0] == [0, 0, 0, 0] # 'This'
        assert sent.indexes['glove'][3] == [2]  # 'a'
        assert sent.indexes['my_char'][0] == [2, 3, 4, 5] # 'This'



import pytest
from antu.io.fields.text_field import TextField


class TestTextField:

    def test_textfield(self):
        sentence = ['This', 'is', 'a', 'test', 'sentence', '.']
        sent = TextField('sentence', sentence)
        assert sent[0] == 'This'
        assert sent[-1] == '.'



import numpy as np

def orthonormal_initializer(output_size, input_size):
    """
    adopted from Timothy Dozat https://github.com/tdozat/Parser/blob/master/lib/linalg.py
    """
    print (output_size, input_size)
    I = np.eye(output_size)
    lr = .1
    eps = .05/(output_size + input_size)
    success = False
    tries = 0
    while not success and tries < 10:
        Q = np.random.randn(input_size, output_size) / np.sqrt(output_size)
        for i in range(100):
            QTQmI = Q.T.dot(Q) - I
            loss = np.sum(QTQmI**2 / 2)
            Q2 = Q**2
            Q -= lr*Q.dot(QTQmI) / (np.abs(Q2 + Q2.sum(axis=0, keepdims=True) + Q2.sum(axis=1, keepdims=True) - 1) + eps)
            if np.max(Q) > 1e6 or loss > 1e6 or not np.isfinite(loss):
                tries += 1
                lr /= 2
                break
        success = True
    if success:
        print('Orthogonal pretrainer loss: %.2e' % loss)
    else:
        print('Orthogonal pretrainer failed, using non-orthogonal random matrix')
        Q = np.random.randn(input_size, output_size) / np.sqrt(output_size)
    return np.transpose(Q.astype(np.float32))


from typing import List, Dict, TypeVar, Callable
from abc import ABCMeta, abstractmethod

from antu.io.vocabulary import Vocabulary

Indices = TypeVar("Indices", List[int], List[List[int]])

class TokenIndexer(metaclass=ABCMeta):
    """
    A ``TokenIndexer`` determines how string tokens get represented as arrays of
    indices in a model.
    """

    @abstractmethod
    def count_vocab_items(
        self,
        token: str,
        counter: Dict[str, Dict[str, int]]) -> None:
        """
        Defines how each token in the field is counted. In most cases, just use
        the string as a key. However, for character-level ``TokenIndexer``, you
        need to traverse each character in the string.

        Parameters
        ----------
        counter : ``Dict[str, Dict[str, int]]``
            We count the number of strings if the string needs to be counted to
            some counters.
        """
        pass

    @abstractmethod
    def tokens_to_indices(
        self,
        tokens: List[str],
        vocab: Vocabulary) -> Dict[str, Indices]:
        """
        Takes a list of tokens and converts them to one or more sets of indices.
        This could be just an ID for each token from the vocabulary.

        Parameters
        ----------
        vocab : ``Vocabulary``
            ``vocab`` is used to get the index of each item.
        """
        pass



from typing import Dict, MutableMapping, Mapping, TypeVar, List

from antu.io.vocabulary import Vocabulary
from antu.io.fields.field import Field

Indices = TypeVar("Indices", List[int], List[List[int]])


class Instance(Mapping[str, Field]):
    """
    An ``Instance`` is a collection (list) of multiple data fields.

    Parameters
    ----------
    fields : ``List[Field]``, optional (default=``None``)
        A list of multiple data fields.
    """

    def __init__(self, fields: List[Field]=None) -> None:
        self.fields = fields
        self.indexed = False  # Indicates whether the instance has been indexed

    def __getitem__(self, key: str) -> Field:
        return self.fields[key]

    def __iter__(self):
        return iter(self.fields)

    def __len__(self) -> int:
        return len(self.fields)

    def add_field(self, field: Field) -> None:
        """
        Add the field to the existing ``Instance``.

        Parameters
        ----------
        field : ``Field``
            Which field needs to be added.
        """
        self.fields.append(field)
        if self.indexed:
            field.index(vocab)

    def count_vocab_items(self, counter: Dict[str, Dict[str, int]]) -> None:
        """
        Increments counts in the given ``counter`` for all of the vocabulary
        items in all of the ``Fields`` in this ``Instance``.

        Parameters
        ----------
        counter : ``Dict[str, Dict[str, int]]``
            We count the number of strings if the string needs to be counted to
            some counters.
        """
        for field in self.fields:
            field.count_vocab_items(counter)

    def index_fields(self, vocab: Vocabulary) -> Dict[str, Dict[str, Indices]]:
        """
        Indexes all fields in this ``Instance`` using the provided ``Vocabulary``.
        This `mutates` the current object, it does not return a new ``Instance``.
        A ``DataIterator`` will call this on each pass through a dataset; we use the ``indexed``
        flag to make sure that indexing only happens once.
        This means that if for some reason you modify your vocabulary after you've
        indexed your instances, you might get unexpected behavior.

        Parameters
        ----------
        vocab : ``Vocabulary``
            ``vocab`` is used to get the index of each item.

        Returns
        -------
        res : ``Dict[str, Dict[str, Indices]]``
            Returns the Indices corresponding to the instance. The first key is
            field name and the second key is the vocabulary name.
        """
        if not self.indexed:
            self.indexed = True
            for field in self.fields:
                field.index(vocab)
        res = {}
        for field in self.fields:
            res[field.name] = field.indexes
        return res




import os
import sys

try:
    from setuptools import setup, find_packages
except:
    from distutils.core import setup

def read_file(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='antu',
    version='0.0.3',
    author='AntNLP',
    author_email='taoji.cs@gmail.com',
    description='Universal data IO and neural network modules in NLP tasks',
    long_description = read_file("README.md"),
    license='Apache',
    packages=find_packages(),
    install_requires=[],
    classifiers = [
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Documentation :: Sphinx',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    url ="https://github.com/AntNLP/antu",
    zip_safe=True,
    include_package_data=True,
    platforms='any',
)



from typing import Dict, List
from abc import ABCMeta, abstractmethod
from antu.io.instance import Instance


class DatasetReader(metaclass=ABCMeta):

    @abstractmethod
    def read(self, file_path: str) -> List[Instance]:
        pass

    @abstractmethod
    def input_to_instance(self, inputs: str) -> Instance:
        pass


import _dynet as dy
import numpy as np

class PointerLabelClassifier(object):
    """This builds Pointer Networks labeled Classifier:
    .. math::
        \\begin{split}
           f_{ptr}(\\boldsymbol{h}_i, \\boldsymbol{s}_t) &=
           \\boldsymbol{V}_a{}^\\top
           \\text{tanh}(\\boldsymbol{W}_1 \\boldsymbol{h}_i +
                       \\boldsymbol{W}_2 \\boldsymbol{s}_t) \\\\
           \\boldsymbol{p}_t &= \\text{softmax}(f_{ptr}(\\boldsymbol{h}_i, \\boldsymbol{s}_t)) \\\\
        \end{split}
    :param model dynet.ParameterCollection:
    :param l_dim int: Row dimension of :math:`\\boldsymbol{V}`
    :param v_dim int: Column dimension of :math:`\\boldsymbol{V}`
    :param h_dim int: Dimension of :math:`\\boldsymbol{h}`
    :param s_dim int: Dimension of :math:`\\boldsymbol{s}`
    :returns: probatilistic vector :math:`\\boldsymbol{p}_t`
    :rtype: dynet.Expression
    """
    def __init__(self, model, l_dim, v_dim, h_dim, s_dim, layers=1):
        pc = model.add_subcollection()
        self.layers = layers
        self.V  = ([pc.add_parameters((1, v_dim)) for _ in range(layers-1)]
                  +[pc.add_parameters((l_dim, v_dim))])

        self.W1 = [pc.add_parameters((v_dim, h_dim)) for _ in range(layers)]

        self.W2 = ([pc.add_parameters((v_dim, s_dim))]
                  +[pc.add_parameters((v_dim, h_dim+s_dim)) for _ in range(layers-1)])

        self.B1 = pc.add_parameters((l_dim, h_dim), init=dy.ConstInitializer(0))
        self.B2 = pc.add_parameters((l_dim, s_dim), init=dy.ConstInitializer(0))

        ## Only single layer support
        #self._W1 = pc.add_parameters((v_dim, h_dim))
        #self._W2 = pc.add_parameters((v_dim, s_dim))
        #self._V  = pc.add_parameters((l_dim, v_dim))

        self.pc = pc
        self.spec = l_dim, v_dim, h_dim, s_dim, layers

    def __call__(self, x, h_matrix, noprob=False):
        s_t = x
        for i in range(self.layers-1):
            e_t = self.V[i] * dy.tanh(self.W1[i]*h_matrix + self.W2[i]*s_t)
            a_t = dy.softmax(dy.transpose(e_t))
            c_t = h_matrix * a_t
            s_t = dy.concatenate([x, c_t])

        e_t = self.V[-1] * dy.tanh(self.W1[-1]*h_matrix + self.W2[-1]*s_t) + self.B1 * h_matrix + self.B2 * s_t

        if len(h_matrix.dim()[0]) > 1:
            e_t = dy.reshape(e_t, (self.V[-1].dim()[0][0] * h_matrix.dim()[0][1],))
        if not noprob:
            p_t = dy.softmax(e_t)
            return p_t
        else:
            return e_t

    @staticmethod
    def from_spec(spec, model):
        """Create and return a new instane with the needed parameters.

        It is one of the prerequisites for Dynet save/load method.
        """
        l_dim, v_dim, h_dim, s_dim, layers = spec
        return PointerLabelClassifier(model, l_dim, v_dim, h_dim, s_dim, layers)

    def param_collection(self):
        """Return a :code:`dynet.ParameterCollection` object with the parameters.

        It is one of the prerequisites for Dynet save/load method.
        """
        return self.pc


class BiaffineLabelClassifier(object):
    """This builds Pointer Networks labeled Classifier:
    .. math::
        \\begin{split}
           f_{ptr}(\\boldsymbol{h}_i, \\boldsymbol{s}_t) &=
           \\boldsymbol{V}_a{}^\\top
           \\text{tanh}(\\boldsymbol{W}_1 \\boldsymbol{h}_i +
                       \\boldsymbol{W}_2 \\boldsymbol{s}_t) \\\\
           \\boldsymbol{p}_t &= \\text{softmax}(f_{ptr}(\\boldsymbol{h}_i, \\boldsymbol{s}_t)) \\\\
        \end{split}
    :param model dynet.ParameterCollection:
    :param l_dim int: Row dimension of :math:`\\boldsymbol{V}`
    :param v_dim int: Column dimension of :math:`\\boldsymbol{V}`
    :param h_dim int: Dimension of :math:`\\boldsymbol{h}`
    :param s_dim int: Dimension of :math:`\\boldsymbol{s}`
    :returns: probatilistic vector :math:`\\boldsymbol{p}_t`
    :rtype: dynet.Expression
    """
    def __init__(self, model, h_dim, s_dim, n_label, h_bias=False, s_bias=False):
        pc = model.add_subcollection()
        if h_bias: h_dim += 1
        if s_bias: s_dim += 1
        if n_label == 1:
            self.U = pc.add_parameters(
                (h_dim, s_dim), init = dy.ConstInitializer(0.))
        else:
            self.U = pc.add_parameters(
                (h_dim*n_label, s_dim), init = dy.ConstInitializer(0.))
        self.pc = pc
        self.h_dim = h_dim
        self.s_dim = s_dim
        self.n_label = n_label
        self.h_bias = h_bias
        self.s_bias = s_bias
        self.spec = (h_dim, s_dim, n_label, h_bias, s_bias)

    def __call__(self, h, s):
        if self.h_bias:
            if len(h.dim()[0]) == 2:
                h = dy.concatenate([h, dy.inputTensor(np.ones((1, h.dim()[0][1]), dtype=np.float32))])
            else:
                h = dy.concatenate([h, dy.inputTensor(np.ones((1,), dtype=np.float32))])
        if self.s_bias:
            if len(s.dim()[0]) == 2:
                s = dy.concatenate([s, dy.inputTensor(np.ones((1, s.dim()[0][1]), dtype=np.float32))])
            else:
                s = dy.concatenate([s, dy.inputTensor(np.ones((1,), dtype=np.float32))])
        lin = self.U * s
        if self.n_label > 1:
            lin = dy.reshape(lin, (self.h_dim, self.n_label))
        blin = dy.transpose(h) * lin
        return blin

    @staticmethod
    def from_spec(spec, model):
        """Create and return a new instane with the needed parameters.

        It is one of the prerequisites for Dynet save/load method.
        """
        h_dim, s_dim, n_label, h_bias, s_bias = spec
        return BiaffineLabelClassifier(model, h_dim, s_dim, n_label, h_bias, s_bias)

    def param_collection(self):
        """Return a :code:`dynet.ParameterCollection` object with the parameters.

        It is one of the prerequisites for Dynet save/load method.
        """
        return self.pc


import _dynet as dy
import numpy as np
from antu.nn.dynet.initializer import orthonormal_initializer


class DeepBiLSTMBuilder(object):
    """This builds deep bidirectional LSTM:

    The original attention mechanism (Bahdanau et al., 2015)
    uses a one-hidden layer feed-forward network to calculate the attention alignment:

    :param model dynet.ParameterCollection:
    :param n_layers int: Number of LSTM layers
    :param x_dim int: Dimension of LSTM input :math:`\\boldsymbol{x}`
    :param h_dim int: Dimension of LSTM hidden state :math:`\\boldsymbol{h}`
    :param LSTMBuilder dynet._RNNBuilder: Dynet LSTM type
    :param param_init bool: Initializes LSTM with parameter
    :returns: (last_output, outputs)
    :rtype: tuple
    """
    def __init__(self, model, n_layers, x_dim, h_dim, LSTMBuilder, param_init=False, fb_fusion=False):
        pc = model.add_subcollection()
        if not fb_fusion:
            self.f = LSTMBuilder(n_layers, x_dim, h_dim, pc)
            self.b = LSTMBuilder(n_layers, x_dim, h_dim, pc)
            if param_init:
                self.f_init = [pc.add_parameters((h_dim),
                               init=dy.ConstInitializer(0))
                               for _ in range(n_layers*2)]
                self.b_init = [pc.add_parameters((h_dim),
                               init=dy.ConstInitializer(0))
                               for _ in range(n_layers*2)]

        else:
            self.DeepBiLSTM = []
            f = LSTMBuilder(1, x_dim, h_dim, pc)
            b = LSTMBuilder(1, x_dim, h_dim, pc)
            self.DeepBiLSTM.append((f, b))
            if n_layers > 1:
                f = LSTMBuilder(n_layers-1, h_dim*2, h_dim, pc)
                b = LSTMBuilder(n_layers-1, h_dim*2, h_dim, pc)
                self.DeepBiLSTM.append((f, b))

        self.param_init = param_init
        self.fb_fusion = fb_fusion
        self.pc = pc
        self.spec = (n_layers, x_dim, h_dim, LSTMBuilder, param_init, fb_fusion)

    def __call__(self, inputs, init_vecs=None, dropout_x = 0., dropout_h = 0., train=False):
        if not self.fb_fusion:
            if self.param_init:
                f, b = self.f.initial_state(self.f_init), self.b.initial_state(self.b_init)
            elif init_vecs:
                f, b = self.f.initial_state(init_vecs["fwd"]), self.b.initial_state(init_vecs["bwd"])
            else:
                f, b = self.f.initial_state(), self.b.initial_state()
            if train:
                self.f.set_dropouts(dropout_x, dropout_h)
                self.b.set_dropouts(dropout_x, dropout_h)
            else:
                self.f.set_dropouts(0., 0.)
                self.b.set_dropouts(0., 0.)
            f_in, b_in = inputs, reversed(inputs)
            f_out, b_out = f.add_inputs(f_in), b.add_inputs(b_in)
            f_last, b_last = f_out[-1].s(), b_out[-1].s()
            f_out, b_out = [state.h()[-1] for state in f_out], [state.h()[-1] for state in b_out]
            out = [dy.concatenate([f, b]) for f, b in zip(f_out, reversed(b_out))]
            last = [dy.concatenate([f, b]) for f, b in zip(f_last, b_last)]
            return (last, out)

        else:
            for f_lstm, b_lstm in self.DeepBiLSTM:
                f, b = f_lstm.initial_state(), b_lstm.initial_state()
                if train:
                    f_lstm.set_dropouts(dropout_x, dropout_h)
                    b_lstm.set_dropouts(dropout_x, dropout_h)
                else:
                    f_lstm.set_dropouts(0., 0.)
                    b_lstm.set_dropouts(0., 0.)
                fs, bs = f.transduce(inputs), b.transduce(reversed(inputs))
                inputs = [dy.concatenate([f,b]) for f, b in zip(fs, reversed(bs))]
            return inputs

    @staticmethod
    def from_spec(spec, model):
        """Create and return a new instane with the needed parameters.

        It is one of the prerequisites for Dynet save/load method.
        """
        n_layers, x_dim, h_dim, LSTMBuilder, param_init, fb_fusion = spec
        return DeepBiLSTMBuilder(model, n_layers, x_dim, h_dim, LSTMBuilder, param_init, fb_fusion)

    def param_collection(self):
        """Return a :code:`dynet.ParameterCollection` object with the parameters.

        It is one of the prerequisites for Dynet save/load method.
        """
        return self.pc

def orthonormal_VanillaLSTMBuilder(n_layers, x_dim, h_dim, pc):
    builder = dy.VanillaLSTMBuilder(n_layers, x_dim, h_dim, pc)
    for layer, params in enumerate(builder.get_parameters()):
        W = orthonormal_initializer(h_dim, h_dim + (h_dim if layer>0 else x_dim))
        W_h, W_x = W[:,:h_dim], W[:,h_dim:]
        params[0].set_value(np.concatenate([W_x]*4, 0))
        params[1].set_value(np.concatenate([W_h]*4, 0))
        b = np.zeros(4*h_dim, dtype=np.float32)
        b[h_dim:2*h_dim] = -1.0
        params[2].set_value(b)
    return builder


from typing import List, Iterator, Dict
from overrides import overrides
from antu.io.token_indexers.token_indexer import TokenIndexer
from antu.io.vocabulary import Vocabulary
from antu.io.fields.field import Field


class SequenceLabelField(Field):

    def __init__(
        self,
        name: str,
        tokens: List[str],
        indexers: List[TokenIndexer]):
        self.name = name
        self.tokens = tokens
        self.indexers = indexers

    def __iter__(self) -> Iterator[str]:
        return iter(self.tokens)

    def __getitem__(self, idx: int) -> str:
        return self.tokens[idx]

    def __len__(self) -> int:
        return len(self.tokens)

    @overrides
    def count_vocab_items(
        self,
        counters: Dict[str, Dict[str, int]]) -> None:
        for idxer in self.indexers:
            for token in self.tokens:
                idxer.count_vocab_items(token, counters)

    @overrides
    def index(
        self,
        vocab: Vocabulary) -> None:
        self.indexes = {}
        for idxer in self.indexers:
            self.indexes.update(idxer.tokens_to_indices(self.tokens, vocab))







import _dynet as dy


class Char2WordCNNEmbedder(object):
    """This builds char to word embedder with CNN:
    :param model dynet.ParameterCollection:
    :param n_char int: Number of char
    :param char_dim int: Dimension of char embedding
    :param n_filter int: Number of CNN filter
    :param win_sizes list: Filter width list
    :returns: c2w_emb
    :rtype: list
    """
    def __init__(self, model, n_char, char_dim, n_filter, win_sizes):
        pc = model.add_subcollection()

        self.clookup = pc.add_lookup_parameters((n_char, char_dim))
        self.Ws = [pc.add_parameters((char_dim, size, 1, n_filter),
                   init=dy.GlorotInitializer(gain=0.5))
                   for size in win_sizes]
        self.bs = [pc.add_parameters((n_filter),
                   init=dy.ConstInitializer(0))
                   for _ in win_sizes]

        self.win_sizes = win_sizes
        self.pc = pc
        self.spec = (n_char, char_dim, n_filter, win_sizes)

    def __call__(self, sentence, c2i, maxn_char, act, train=False):
        words_batch = []
        for token in sentence:
            chars_emb = [self.clookup[int(c2i.get(c, 0))] for c in token.chars]
            c2w = dy.concatenate_cols(chars_emb)
            c2w = dy.reshape(c2w, tuple(list(c2w.dim()[0]) + [1]))
            words_batch.append(c2w)

        words_batch = dy.concatenate_to_batch(words_batch)
        convds = [dy.conv2d(words_batch, W, stride=(1, 1), is_valid=True) for W in self.Ws]
        actds = [act(convd) for convd in convds]
        poolds = [dy.maxpooling2d(actd, ksize=(1, maxn_char-win_size+1), stride=(1, 1))
                  for win_size, actd in zip(self.win_sizes, actds)]
        words_batch = [dy.reshape(poold, (poold.dim()[0][2],)) for poold in poolds]
        words_batch = dy.concatenate([out for out in words_batch])

        c2w_emb = []
        for idx, token in enumerate(sentence):
            c2w_emb.append(dy.pick_batch_elem(words_batch, idx))
        return c2w_emb

    @staticmethod
    def from_spec(spec, model):
        """Create and return a new instane with the needed parameters.

        It is one of the prerequisites for Dynet save/load method.
        """
        n_char, char_dim, n_filter, win_sizes = spec
        return Char2WordCNNEmbedder(model, n_char, char_dim, n_filter, win_sizes)

    def param_collection(self):
        """Return a :code:`dynet.ParameterCollection` object with the parameters.

        It is one of the prerequisites for Dynet save/load method.
        """
        return self.pc